//+------------------------------------------------------------------+
//|                                                      TradeNN.mq5 |
//|                                    Copyright (c) 2023, Marketeer |
//|                          https://www.mql5.com/ru/articles/12187/ |
//+------------------------------------------------------------------+
#include <AutoPtr.mqh>
#define BATCH_PROP
#include <MatrixNet.mqh>
#include <MatrixNetStore.mqh>
//#define SKIP_EXPERIMENTS

input string Symbols = "XAGUSD,XAUUSD,EURUSD";;
input int Depth = 5; // Vector size (bars)
input int Reserve = 250; // Training set size (vectors)
input int Epochs = 1000;
input double Accuracy = 0.0001; // Accuracy (and training speed)
input double HiddenLayerFactor = 2.0; // Hidden Layers Factor (to vector size)
input int DropOutPercentage = 0; // DropOut Percentage
#ifdef BATCH_PROP
input double SpeedUp = 1.1;
input double SpeedDown = 0.1;
input double SpeedHigh = 50;
input double SpeedLow = 0;
#endif
input string NetBinFileName = "";
input int Randomizer = 0;

string S[];
int Q;
string FrameFile;
double FrameLoss;

//+------------------------------------------------------------------+
//| Array of doubles holder (used for Close proces here)             |
//+------------------------------------------------------------------+
struct Closes
{
   double C[];
};

//+------------------------------------------------------------------+
//| Array of arrays of doubles (used for quotes of a set of symbols) |
//+------------------------------------------------------------------+
Closes CC[];

/*
              TRAINING
                               incomplete (doesn't have close price), not used
              4   3   2   1    0  <- bars
   XAUUSD  ...  +   +   *         <- deltas ('*' - "future" target, '+' - known history)
   XAGUSD  ...  +   +   -         <- ('-' - "future" bars on other symbols,
   XPTUSD  ...  +   +   -         <- could be predicted with multi-output NN, not used here)
   
              PREDICTION

              4   3   2   1   0
   XAUUSD      ...  +   +   *
   XAGUSD      ...  +   +   -
   XPTUSD      ...  +   +   -
              
   Only the 0-th symbol has 1 additional diff-value for target in training
*/

//+------------------------------------------------------------------+
//| Price changes per bar for all work symbols, for required history |
//+------------------------------------------------------------------+
bool Calc(const int offset)
{
   // when offset is 0, and new bar is just created,
   // the 0-th bar has almost the same price as the 1-st:
   // this empty difference is expected for current symbol, because it's to be predicted
   const datetime dt = iTime(_Symbol, _Period, offset);
   for(int i = 0; i < Q; ++i)
   {
      const int bar = iBarShift(S[i], PERIOD_CURRENT, dt);
      // +1 for delta, +1 for model/target (accuracy check, on how predictable the symbol/TF is)
      const int n = CopyClose(S[i], PERIOD_CURRENT,
         bar, Depth + 2, CC[i].C);
      if(n != Depth + 2)
      {
         PrintFormat("No data for %s at %s[%d]", S[i], (string)dt, offset);
         return false;
      }
      
      for(int j = 0; j < n - 1; ++j)
      {
         CC[i].C[j] = (CC[i].C[j + 1] - CC[i].C[j]) /
            SymbolInfoDouble(S[i], SYMBOL_TRADE_TICK_SIZE) * SymbolInfoDouble(S[i], SYMBOL_TRADE_TICK_VALUE);
      }
      
      ArrayResize(CC[i].C, n - 1);
   }
   
   return true;
}

//+------------------------------------------------------------------+
//| Price action in bar sequences for all work symbols               |
//+------------------------------------------------------------------+
double Diff(const double &a[], double &d[], const bool actual = false)
{
   const int n = ArraySize(a);
   ArrayResize(d, n - 1); // -1 to eliminate "future" model value, (n - 1) should be equal to Depth
   double overall = 0;
   for(int j = 0; j < n - 1; ++j) // from left (old) to the right (new)
   {
      int k = n - 2 - j;
      overall += a[k];
      d[j] = overall / sqrt(j + 1);
   }

   // preliminary sub-vector normalization per symbol
   // may boost training efficiency, but experimental
   #ifndef SKIP_EXPERIMENTS
   vector temp;
   temp.Assign(d); // should be replaced with Swap
   double norm = temp.Norm(VECTOR_NORM_P, 2);
   // estimate of average price volatility in money
   if(norm == 0) norm = 1.0;
   temp /= norm;
   temp.Swap(d);
   // last element is returned on its own (apart from the array) and
   // will be used as a target (for current symbol) or get skipped (for other symbols)
   double result = actual ? a[n - 1] : a[n - 1] / norm;
   return result;
   #else
   return a[n - 1];
   #endif
}

//+------------------------------------------------------------------+
//| Program start-up                                                 |
//+------------------------------------------------------------------+
int OnInit()
{
   // TODO: check period for D1+ and show a warning if less
   
   Q = StringSplit(StringLen(Symbols) ? Symbols : _Symbol, ',', S);
   if(S[0] != _Symbol)
   {
      Alert(StringFormat("Symbols list should start with current chart symbol (%s), %s given", _Symbol, S[0]));
      return INIT_FAILED;
   }
   ArrayResize(CC, Q);
   MathSrand(Randomizer);
   EventSetTimer(1);
   
   return INIT_SUCCEEDED;
}

//+------------------------------------------------------------------+
//| Timer handler                                                    |
//+------------------------------------------------------------------+
void OnTimer()
{
   EventKillTimer();
   if(!MQLInfoInteger(MQL_TESTER))
   {
      OnTick();
   }
}

//+------------------------------------------------------------------+
//| NN loading from given file                                       |
//| we're loading from FILE_COMMON by default, because otherwise     |
//| we need to hardcode filename in the source for the tester        |
//| to pickup it from #property tester_file                          |
//+------------------------------------------------------------------+
MatrixNet *LoadNet(const string filename, double &std, const int flags = FILE_COMMON)
{
   BinFileNetStorage reader; // custom (optional) data
   MatrixNetStore store;     // general data
   MatrixNet *net;
   std = 1.0;
   Print("Loading ", filename);
   ResetLastError();
   net = store.load<MatrixNet>(filename, &reader, flags);
   if(net == NULL)
   {
      Print("Failed: ", _LastError);
      return NULL;
   }
   MatrixNet::Stats s[1];
   s[0] = reader.getStats();
   ArrayPrint(s);
   std = reader.getScale();
   Print(std);
   Print(reader.getDescription());
   return net;
}

//+------------------------------------------------------------------+
//| Trading by signals of the given NN                               |
//+------------------------------------------------------------------+
bool TradeTest(MatrixNet *net, const double std)
{
   if(!Calc(0)) return false;
   double coefs[];
   for(int i = 0; i < Q; ++i)
   {
      double temp[];
      /* double m = */Diff(CC[i].C, temp, true); // difference for 0-th bar is empty and ignored
      ArrayCopy(coefs, temp, ArraySize(coefs), 0);
   }
    
   vector t;
   t.Assign(coefs);
    
   matrix data = {};
   data.Row(t, 0);
   data /= std;
   
   ClosePosition();
   
   if(net.feedForward(data))
   {
      matrix y = net.getResults();
      Print("Prediction: ", y[0][0] * std);
      
      OpenPosition((y[0][0] > 0) ? ORDER_TYPE_BUY : ORDER_TYPE_SELL);
      return true;
   }
   return false;
}

//+------------------------------------------------------------------+
//| Helper function to start new trade                               |
//+------------------------------------------------------------------+
bool OpenPosition(const ENUM_ORDER_TYPE type)
{
   // define empty struct
   MqlTradeRequest request = {};
   
   // fill required values
   request.action = TRADE_ACTION_DEAL;
   request.type = type;
   request.symbol = _Symbol;
   request.volume = SymbolInfoDouble(_Symbol, SYMBOL_VOLUME_MIN);
   request.price = SymbolInfoDouble(_Symbol, type == ORDER_TYPE_BUY ? SYMBOL_ASK : SYMBOL_BID);
   request.deviation = 5;

   // fill optional fields
   // request.magic = Magic;
   
   ResetLastError();
   MqlTradeResult result[1];
   const bool ok = OrderSend(request, result[0]);
   
   Print("Status: ", _LastError);
   ArrayPrint(result);
   
   if(ok && (result[0].retcode == TRADE_RETCODE_DONE
          || result[0].retcode == TRADE_RETCODE_PLACED))
   {
      return true;
   }

   return false;
}

//+------------------------------------------------------------------+
//| Helper function to close trade                                   |
//+------------------------------------------------------------------+
bool ClosePosition()
{
   // define empty struct
   MqlTradeRequest request = {};

   if(!PositionSelect(_Symbol)) return false;
   const string pl = StringFormat("%+.2f", PositionGetDouble(POSITION_PROFIT));
   
   // fill required values
   request.action = TRADE_ACTION_DEAL;
   request.position = PositionGetInteger(POSITION_TICKET);
   const ENUM_ORDER_TYPE type = (ENUM_ORDER_TYPE)(PositionGetInteger(POSITION_TYPE) ^ 1);
   request.type = type;
   request.price = SymbolInfoDouble(_Symbol, type == ORDER_TYPE_BUY ? SYMBOL_ASK : SYMBOL_BID);
   request.volume = PositionGetDouble(POSITION_VOLUME);
   request.deviation = 5;
   request.comment = pl;
   
   // fill optional fields of the struct
   // request.magic = Magic;

   ResetLastError();
   MqlTradeResult result[1];
   const bool ok = OrderSend(request, result[0]);
   
   Print("Status: ", _LastError, ", P/L: ", pl);
   ArrayPrint(result);
   
   if(ok && (result[0].retcode == TRADE_RETCODE_DONE
          || result[0].retcode == TRADE_RETCODE_PLACED))
   {
      return true;
   }

   return false;
}

//+------------------------------------------------------------------+
//| Helper function to wait until all work symbols bars are in-sync  |
//+------------------------------------------------------------------+
bool Sync()
{
   const datetime dt = iTime(_Symbol, _Period, 0);
   for(int i = 1; i < Q; ++i)
   {
      if(iTime(S[i], _Period, 0) != dt)
      {
         return false;
      }
   }
   return true;
}

MatrixNet *TrainNet(double &std)
{
   double coefs[];
   matrix sys(Reserve, Q * Depth);
   vector model(Reserve);
   vector t;
   datetime start = 0;
  
   for(int j = Reserve - 1; j >= 0; --j)
   {
      // since we're using close price, make +1 in offset
      if(!Calc(j + 1)) // get data for all symbols starting from j-th bar backward on Depth bars
      {
         return NULL; // probably some other symbols may not have enough history
      }
    
      // remember where/when the training set begins
      if(start == 0) start = iTime(_Symbol, _Period, j);
    
      ArrayResize(coefs, 0);
    
      // get increments from plain prices for every symbol on Depth bars
      for(int i = 0; i < Q; ++i)
      {
         double temp[];
         double m = Diff(CC[i].C, temp);
         if(i == 0)
         {
            model[j] = m;
         }
         int dest = ArraySize(coefs);
         ArrayCopy(coefs, temp, dest, 0);
      }
    
      t.Assign(coefs);
      sys.Row(t, j);
   }

   std = sys.Std() * 3;
   Print("Normalization by 3 std: ", std);
   sys /= std;

   matrix target = {};
   target.Col(model, 0);
   target /= std;
  
   // sizes of layers 0, 1, 2, 3 are deduced from data, output is the one
   int layers[] = {0, 0, 0, 0, 1};
   layers[0] = (int)sys.Cols();
   layers[1] = (int)(sys.Cols() * HiddenLayerFactor);
   layers[2] = (int)(sys.Cols() * HiddenLayerFactor);
   layers[3] = (int)fmax(sqrt(sys.Rows()), fmax(sqrt(layers[1] * layers[3]), sys.Cols() * sqrt(HiddenLayerFactor)));

   ArrayPrint(layers);
   MatrixNetVisual *net = new MatrixNetVisual(layers);
   #ifdef BATCH_PROP
   net.setupSpeedAdjustment(SpeedUp, SpeedDown, SpeedHigh, SpeedLow);
   #endif
   net.enableDropOut(DropOutPercentage);

   Print("Training result: ", net.train(sys, target, Epochs, Accuracy));
   
   matrix w[];
   if(net.getBestWeights(w))
   {
      MatrixNet net2(w); // optionally can customize activation functions
      if(net2.isReady())
      {
         Print("Best result: ", net2.test(sys, target));
         
         // most important or all expert settings can be stored in the net file
         const string context = StringFormat("\r\n%s %s %s-%s", _Symbol, EnumToString(_Period),
            TimeToString(start), TimeToString(iTime(_Symbol, _Period, 0))) + "\r\n" +
            Symbols + "\r\n" + (string)Depth + "/" + (string)Reserve + "\r\n" +
            (string)Epochs + "/" + (string)Accuracy + "\r\n" +
            (string)HiddenLayerFactor + "/" + (string)DropOutPercentage + "\r\n";
            
         // prepare temp name for the file
         const string tempfile = "bpnnmtmp" + (string)GetTickCount64() + ".bpn";

         // save the net and custom data into the file
         MatrixNetStore store;
         BinFileNetStorage writer(context, net.getStats(), std);
         store.save(tempfile, *net, &writer);

         // during MQLInfoInteger(MQL_OPTIMIZATION) we'll discover
         // which parameters are optimized, and include their values into the file name
         
         if(!MQLInfoInteger(MQL_OPTIMIZATION))
         {
            // rename the file to more user-friendly timestamp, in common folder
            string filename = "bpnnm" + TimeStamp((datetime)FileGetInteger(tempfile, FILE_MODIFY_DATE))
               + StringFormat("(%7g)", net.getStats().bestLoss) + ".bpn";
            if(!FileMove(tempfile, 0, filename, FILE_COMMON))
            {
               PrintFormat("Can't rename temp-file: %s [%d]", tempfile, _LastError);
               filename = tempfile;
            }
   
            Print("Check-up of saved and restored copy: ", filename);
            MatrixNet *net3 = LoadNet(filename, std);
            if(net3)
            {
               Print("Best result restored: ", net3.test(sys, target));
               delete net3;
            }
         }
         else // the file should be sent from the agent to MT5 via frame
         {
            FrameFile = tempfile;
            FrameLoss = net.getStats().bestLoss;
         }
      }
   }
   return net;
}

#define WAIT 5000
//+------------------------------------------------------------------+
//| Main program handler                                             |
//+------------------------------------------------------------------+
void OnTick()
{
   // we must wait for new bars on all work symbols (if many)
   int start = 0;
   while(!Sync())
   {
      start += WAIT;
      if(start > PeriodSeconds() / 2) return;
      Sleep(WAIT);
   }

   // NB: static variables (and EA itself) are not re-setted on symbol/timeframe change!
   // this is a demo EA for tester

   // when no existing network provided,
   // wait for sufficient bars to train a new one
   if(NetBinFileName == "")
   {
      static bool warning1 = false;
      static bool warning2 = false;
      if(Reserve + Depth + 2 > iBars(_Symbol, _Period))
      {
         if(!warning1)
         {
            PrintFormat("Insufficient bars: %d", iBars(_Symbol, _Period));
            warning1 = true;
         }
         return;
      }
      else if(!warning2)
      {
         Print("Sufficient bars at: ", iTime(_Symbol, _Period, 0));
         warning2 = true;
      }
   }

   // actual work goes on "once per bar" basis
   static datetime last = 0;
   if(last == iTime(_Symbol, _Period, 0)) return;
  
   static AutoPtr<MatrixNet> run;
   static double std;

   if(NetBinFileName != "")
   {
      if(!run[])
      {
         run = LoadNet(NetBinFileName, std);
         if(!run[])
         {
            ExpertRemove();
            return;
         }
         // TODO: on success wait for new bar if not in the tester
      }
   }
   
   if(run[])
   {
      TradeTest(run[], std);
   }
   else
   {
      run = TrainNet(std);      
   }
   
   last = iTime(_Symbol, _Period, 0);
}

//+------------------------------------------------------------------+
//| Custom writer/reader for NN meta-data                            |
//+------------------------------------------------------------------+
class BinFileNetStorage: public Storage
{
protected:
   // custom data processing
   string description;
   double normalizationScale;      // Standard Deviation here
   // const int normalizationType; // always 0 yet (only one type supported)
   MatrixNet::Stats stats;
   
public:
   BinFileNetStorage(const string about, const MatrixNet::Stats &s, const double stddev)
   {
      description = about;
      stats = s;
      normalizationScale = stddev;
   }
   
   BinFileNetStorage() { }
   
   MatrixNet::Stats getStats() const
   {
      return stats;
   }
   
   double getScale() const
   {
      return normalizationScale;
   }
   
   string getDescription() const
   {
      return description;
   }
   
   virtual bool store(const int h) override
   {
      if(FileGetInteger(h, FILE_IS_BINARY) <= 0) return false;
      
      FileWriteInteger(h, StringLen(description));
      FileWriteString(h, description);
      FileWriteDouble(h, normalizationScale);
      return FileWriteStruct(h, stats) > 0;
   }

   virtual bool restore(const int h) override
   {
      if(FileGetInteger(h, FILE_IS_BINARY) <= 0) return false;

      const int len = FileReadInteger(h);
      description = FileReadString(h, len);
      normalizationScale = FileReadDouble(h);
      return FileReadStruct(h, stats) > 0;
   }
};

//+------------------------------------------------------------------+
//| YYYYMMDDHHmm                                                     |
//+------------------------------------------------------------------+
string TimeStamp(const datetime dt = LONG_MAX, const int mode = TIME_DATE | TIME_MINUTES)
{
   string now = TimeToString(dt == LONG_MAX ? TimeCurrent() : dt, mode);
   StringReplace(now, ".", "");
   StringReplace(now, ":", "");
   StringReplace(now, " ", "");
   return now;
}

//+------------------------------------------------------------------+
//| Tester event handler for sending frame (dummy custom estimator)  |
//+------------------------------------------------------------------+
double OnTester()
{
   return FrameAdd("BPNNFILE", 0, FrameLoss, FrameFile);
}

//+------------------------------------------------------------------+
//| Frame with NN file to save into a file                           |
//+------------------------------------------------------------------+
void ProcessFileFrames()
{
   ulong pass;
   string name;
   long id;
   double result;
   uchar data[];
  
   while(FrameNext(pass, name, id, result, data))
   {
      string params[];
      uint count;
      if(FrameInputs(pass, params, count))
      {
         string record = MQLInfoString(MQL_PROGRAM_NAME) + "/" + TimeStamp(TimeLocal(), TIME_DATE) + "/";
         StringReplace(record, ".mq5/", "/");
         for(uint i = 0; i < count; i++)
         {
            string name2value[];
            int n = StringSplit(params[i], '=', name2value);
            if(n == 2)
            {
               long pvalue, pstart, pstep, pstop;
               bool enabled = false;
               if(ParameterGetRange(name2value[0], enabled, pvalue, pstart, pstep, pstop))
               {
                  if(enabled)
                  {
                     record += params[i] + "-";
                  }
               }
            }
         }

         const string filename = record + StringFormat("%s(%7g).bpn", (string)pass, result);
         int h = FileOpen(filename, FILE_WRITE | FILE_ANSI | FILE_BIN);
         if(h != INVALID_HANDLE)
         {
            FileWriteArray(h, data);
            FileClose(h);
         }
         else
         {
            PrintFormat("Can't save %s [%d]", filename, _LastError);
         }
      }
   }
}

//+------------------------------------------------------------------+
//| Tester manager event handler                                     |
//+------------------------------------------------------------------+
void OnTesterPass()
{
   ProcessFileFrames();
}

//+------------------------------------------------------------------+
//| Tester manager final event handler                               |
//+------------------------------------------------------------------+
void OnTesterDeinit()
{
   ProcessFileFrames();
}

//+------------------------------------------------------------------+
